#include "fsplugin.h"
#include "cunicode.h"
#ifdef _WIN32
# include <shlwapi.h>	// wnsprinf()
# define vsnprintf _vsnprintf
# define EXPORT(ret) extern "C" ret __declspec(dllexport) WINAPI
# define strcasecmp _stricmp
#else
# define EXPORT(ret) __attribute__((visibility(standard))) extern "C" ret
#endif
#include "Xml.h"


HINSTANCE hInstance;
char inifilename[MAX_PATH];  // Unused in this plugin, may be used to save data
volatile bool wantStop;

static void _cdecl debug(char*t,...) {
#ifdef _DEBUG
 char buf[1024];
 va_list va;
 va_start(va,t);
 vsnprintf(buf,elemof(buf),t,va);
 OutputDebugStringA(buf);
 va_end(va);
#endif
}

extern "C" BOOL APIENTRY _DllMainCRTStartup(HINSTANCE h,DWORD reason,void*) {
 debug("DllMain(%p,%d)\n",h,reason);
 switch (reason) {
  case DLL_PROCESS_ATTACH: {
   hInstance=h;
   DisableThreadLibraryCalls(h);
  }break;
 }
 return TRUE;
}

int PluginNumber;
tProgressProc ProgressProc;
tLogProc LogProc;
tRequestProc RequestProc;
tProgressProcW ProgressProcW;
tLogProcW LogProcW;
tRequestProcW RequestProcW;

EXPORT(int) FsInit(int n,tProgressProc p1,tLogProc p2,tRequestProc p3) {
 PluginNumber=n;
 ProgressProc=p1;
 LogProc=p2;
 RequestProc=p3;
 return 0;
}

EXPORT(int) FsInitW(int n,tProgressProcW p1,tLogProcW p2,tRequestProcW p3) {
 debug("%s(%d,%p,%p,%p)\n","FsInitW",n,p1,p2,p3);
 PluginNumber=n;
 ProgressProcW=p1;
 LogProcW=p2;
 RequestProcW=p3;
 return 0;
}

struct FindHandle{
 int iItem;
 const fs::Entry*item;
 operator HANDLE() {return HANDLE(this);}
 FindHandle(const fs::Entry*n):item(n),iItem(0) {}
};

struct LoadedFile{
 Xml xml;	// .hHeap!=0 when loaded
 int dirty;	// counts number of changes
 DWORD packer_id;
 FILETIME getLastMod() const;
 char displayname[64];	// Katalogname aus XML-Datei
 char filename[MAX_PATH];
 bool load();
 bool save();
 ~LoadedFile();
 static bool convertFile(HANDLE,HANDLE,const char*);
 static bool process(HANDLE&,const char*,bool=false);
private:
 HANDLE open(DWORD creation=OPEN_EXISTING) const;
};

LoadedFile::~LoadedFile() {
 if (xml.hHeap) {
  if (dirty) save();
  HeapDestroy(xml.hHeap);
  xml.hHeap=0;
 }
}

HANDLE LoadedFile::open(DWORD creation) const{
 int l=MultiByteToWideChar(CP_UTF8,0,filename,-1,0,0);
 WCHAR*n=(WCHAR*)_alloca(l*sizeof*n);
 MultiByteToWideChar(CP_UTF8,0,filename,-1,n,l);
 return CreateFileW(n,
	creation==OPEN_EXISTING?GENERIC_READ:GENERIC_READ|GENERIC_WRITE,
	FILE_SHARE_READ,0,creation,0,0);
}

FILETIME LoadedFile::getLastMod() const{
 union{
  ULONGLONG ull;
  FILETIME ft;
 }ret={-1};
 HANDLE h=open();
 if (h!=INVALID_HANDLE_VALUE) {
  GetFileTime(h,0,0,&ret.ft);
  CloseHandle(h);
 }
 return ret.ft;
}

bool LoadedFile::convertFile(HANDLE hr,HANDLE hw,const char*cmd) {
 int l=MultiByteToWideChar(CP_UTF8,0,cmd,-1,0,0);
 WCHAR*p=(WCHAR*)_alloca(l*sizeof*p);
 MultiByteToWideChar(CP_UTF8,0,cmd,-1,p,l);
 STARTUPINFOW si;
 memset(&si,0,sizeof si);
 si.cb=sizeof si;
 si.dwFlags=STARTF_USESTDHANDLES|STARTF_FORCEOFFFEEDBACK;
// SetHandleInformation(hr,HANDLE_FLAG_INHERIT,HANDLE_FLAG_INHERIT);
// SetHandleInformation(hw,HANDLE_FLAG_INHERIT,HANDLE_FLAG_INHERIT);
 si.hStdInput=hr;
 si.hStdOutput=hw;
 si.hStdError=hw;
 PROCESS_INFORMATION pi;
 if (!CreateProcessW(0,p,0,0,TRUE,CREATE_NO_WINDOW,0,0,&si,&pi)) return false;
 CloseHandle(pi.hThread);
 CloseHandle(pi.hProcess);
 CloseHandle(hr);	// CreateProcess ruft intern DuplicateHandle()
 CloseHandle(hw);
 return true;
}


// Packt Datei im Worker-Thread aus und tauscht das Dateihandle gegen das Read-Pipe-Handle
// Oder packt Datei im Worker-Thread ein und tauscht das Dateihandle gegen das Write-Pipe-Handle
bool LoadedFile::process(HANDLE&h,const char*cmd,bool compress) {
 HANDLE hr,hw;
 if (!CreatePipe(&hr,&hw,0,Xml::BUFSIZE)) return false;
 if (compress) {
  if (!convertFile(hr,h,cmd)) return false;
  h=hw;
 }else{
  if (!convertFile(h,hw,cmd)) return false;
  h=hr;
 }
 return true;
}

bool LoadedFile::load() {
 if (xml.hHeap) return true;	// bereits geladen
 HANDLE h=open();
 if (h==INVALID_HANDLE_VALUE) return false;
 debug("load %s start\n",filename);
 DWORD br;
 ReadFile(h,&packer_id,4,&br,0);
 SetFilePointer(h,0,0,FILE_BEGIN);
// Kann eine Weile dauern! ProgressProcW() aufrufen
 WCHAR f[256],s[64];
 MultiByteToWideChar(CP_UTF8,0,filename,-1,f,elemof(f));
 LoadStringW(hInstance,3,s,elemof(s));	// "RAM-Speicher"
 ProgressProcW(PluginNumber,f,s,0);
 bool ret=true;
// unpackFile() geht schnell (blockiert kaum),
// denn das ersetzt <h> durch eine Pipe mit einem Ftter-Thread
 switch (packer_id) {
  case 0x00088B1F: ret=process(h,"gunzip"); break;
  case 0x31685A42: ret=process(h,"bunzip2"); break;
 }
 if (ret) {
  xml.hHeap=HeapCreate(HEAP_GENERATE_EXCEPTIONS|HEAP_NO_SERIALIZE,0,0);
  ret=xml.parse(h);	// DAS dauert lange! Lsst sich nur schwerlich parallelisieren.
 }
 ProgressProcW(PluginNumber,f,s,100);
 debug("load %s %s\n",filename,ret?"success":"fail");
 if (!ret) debug("Error %x at %s:%u:%u\n",filename,xml.ec,xml.line+1,xml.pos+1);
 return ret;
}

bool LoadedFile::save() {
// make backup!!
 HANDLE h=open(CREATE_ALWAYS);
 if (h==INVALID_HANDLE_VALUE) return false;
 switch (packer_id) {
  case 0x00088B1F: if (!process(h,"gzip",true)) return false; break;
  case 0x31685A42: if (!process(h,"bzip2",true)) return false; break;
 }
 return xml.write(h);
}

// std::vector beit sich mit /NOD (nodefaultlib)
class LoadedFiles{
 int fill;
 LoadedFile a[10];
public:
 const LoadedFile&operator[](int i) const {return a[i];}
 LoadedFile&operator[](int i) {return a[i];}
 int size() const {return fill;}
 void size(int s) {fill=s;}
}loadedFiles;

byte classify(const fs::Entry*n,byte check=0xFF) {	// Liefert 1-aus-n-Bitmaske
 return n && n->type&check;
}

static fs::Entry*pathToNode(const WCHAR*s,byte allow=eMedia|eDir,char*name=0) {
 if (s[0]!='\\') return 0;	// Nur absoluter Pfad erlaubt!
 if (!s[1]) return 0;		// Wurzel = Liste der Dateinamen
 int len=WideCharToMultiByte(CP_UTF8,0,s+1,-1,0,0,0,0);
 char*p=(char*)_alloca(len+2);
 WideCharToMultiByte(CP_UTF8,0,s+1,-1,p,len,0,0);
 p[len]=0; p[len+1]=0;
 for (char*q=p;;++q) {
  q=strchr(q,'\\');
  if (q) *q=0; else break;	// Doppelnulltertimierten Pfadstring generieren
 }
// Wenn Namenszeiger <name> gegeben, dann letzte Pfadkomponente abspalten
 if (name) {
  if (!*p) return 0;		// Keine Pfadkomponente = Fehler
  char*q=p;
 }
 for (int i=0; i<loadedFiles.size(); i++) {
  if (!strcmp(loadedFiles[i].displayname,p)) {	// Dateiname gefunden
   if (!loadedFiles[i].load()) return 0;	// Kann nicht laden
   fs::Entry*n=loadedFiles[i].xml.root.sub;
   for(;n;n=n->next) if (classify(n,eCatalog)) break;
   if (!n) continue;	// Kein Katalog
   for (char*q=p; *(q+=strlen(q)+1);) {
    for(n=n->sub;n;n=n->next) if (n->type&allow && !strcmp(n->name,q)) break;
    if (!n) return 0;	// Kein passender Pfad
   }
   return n;
  }
 }
 return 0;
}

static const fs::Entry*setFindData(const fs::Entry*c,WIN32_FIND_DATAW&W) {
 for (;c;c=c->next) if (classify(c,eMedia|eDir|eFile)) break;	// Kommentare u.a. Metainfo berspringen
 if (!c) return c;
 if (c->ftLastMod) {
  FILETIME ft;
  DosDateTimeToFileTime(HIWORD(c->ftLastMod),LOWORD(c->ftLastMod),&ft);
  LocalFileTimeToFileTime(&ft,&W.ftLastWriteTime);
 }
 if (c->type==eFile) {	// Datei, kein Verzeichnis
  W.nFileSizeHigh=DWORD(c->size>>32);
  W.nFileSizeLow=DWORD(c->size);
 }else W.dwFileAttributes|=FILE_ATTRIBUTE_DIRECTORY;
 unicopy(W.cFileName,elemof(W.cFileName),c->name);
 return c;
}

static void initfinddata(WIN32_FIND_DATAW&W) {
 memset(&W,0,sizeof W);
 W.ftLastWriteTime.dwHighDateTime=DWORD(-1);
 W.ftLastWriteTime.dwLowDateTime=DWORD(-2);
}

//static bool classify1(const Xml::Node*n,void*) {
// return !!classify(n,/*eCatalog|*/eMedia|eDir|eFile);
//}

EXPORT(HANDLE) FsFindFirstW(const WCHAR*Path,WIN32_FIND_DATAW&W) {
// debug("%s(%S,%p)\n","FsFindFirstW",Path,&W);
 initfinddata(W);
 if (Path[0]=='\\' && !Path[1]) {
  W.dwFileAttributes=FILE_ATTRIBUTE_READONLY|FILE_ATTRIBUTE_REPARSE_POINT;		// Kein Verzeichnis zum ffnen
  FindHandle*lf=new FindHandle(0);
  LoadStringW(hInstance,2,W.cFileName,elemof(W.cFileName));	//"<Neuer Katalog>"
  return lf;
 }
 const fs::Entry*n=pathToNode(Path);
 if (!classify(n,eCatalog|eMedia|eDir)) return INVALID_HANDLE_VALUE;
 const fs::Entry*c=setFindData(n->sub,W);
 if (!c) {
  SetLastError(ERROR_NO_MORE_FILES);
  return INVALID_HANDLE_VALUE;
 }
 return new FindHandle(c);
}

EXPORT(HANDLE) FsFindFirst(const char*Path,WIN32_FIND_DATA&A) {
 WIN32_FIND_DATAW W;
 WCHAR PathW[LONGPATH_MAX];
 HANDLE retval=FsFindFirstW(unicopy(PathW,elemof(PathW),Path),W);
 if (retval!=INVALID_HANDLE_VALUE) unicopy(A,W);
 return retval;
}

EXPORT(BOOL) FsFindNextW(HANDLE Hdl,WIN32_FIND_DATAW&W) {
// debug("%s(%p,%p)\n","FsFindNextW",Hdl,&W);
 initfinddata(W);
 FindHandle*fh=(FindHandle*)Hdl;
 if (!fh->item) {   // generate loaded files list
  if (++fh->iItem>loadedFiles.size()) {
   SetLastError(ERROR_NO_MORE_FILES);
   return false;
  }
  W.dwFileAttributes=FILE_ATTRIBUTE_DIRECTORY;
  LoadedFile&lf=loadedFiles[fh->iItem-1];
  W.ftLastWriteTime=lf.getLastMod();
  unicopy(W.cFileName,elemof(W.cFileName),lf.displayname);
  return true;
 }
 ++fh->iItem;
 if (!(fh->item=setFindData(fh->item->next,W))) {
  SetLastError(ERROR_NO_MORE_FILES);
  return false;
 }
 return true;
}

EXPORT(BOOL) FsFindNext(HANDLE Hdl,WIN32_FIND_DATA&A) {
 WIN32_FIND_DATAW W;
 unicopy(W,A);
 BOOL retval=FsFindNextW(Hdl,W);
 if (retval) unicopy(A,W);
 return retval;
}

EXPORT(int) FsFindClose(HANDLE Hdl) {
// debug("%s(%p)\n","FsFindClose",Hdl);
 FindHandle*fh=(FindHandle*)Hdl;
 delete fh;
 return 0;
}

EXPORT(void) FsGetDefRootName(char*DefRootName,int maxlen) {
 debug("%s(%p,%d)\n","FsGetDefRootName",DefRootName,maxlen);
 LoadString(hInstance,1,DefRootName,maxlen);	// "CD-Katalog"
}

EXPORT(BOOL) FsMkDir(const char*Path) {
 WCHAR wbuf[260];
 return FsMkDirW(unicopy(wbuf,elemof(wbuf),Path));
}

EXPORT(BOOL) FsMkDirW(const WCHAR*Path) {
 char name[260];
 fs::Entry*e=pathToNode(Path,eMedia|eDir,name);
 if (!e) return FALSE;
 const Xml*x=Xml::locate(e);
 if (!x) return FALSE;
 fs::Entry*n=x->newNode(e);	// an der Kindliste von <e> anhngen
 n->type=eDir;
 n->name=x->dup(name);
 return TRUE;
}
#if 0

EXPORT(int) FsRenMovFile(const char*OldName,const char*NewName,BOOL Move,BOOL OverWrite,RemoteInfoStruct*ri) {
 WCHAR OldNameW[LONGPATH_MAX],NewNameW[LONGPATH_MAX];
 return FsRenMovFileW(
	unicopy(OldNameW,elemof(OldNameW),OldName),
	unicopy(NewNameW,elemof(NewNameW),NewName),
	Move,OverWrite,ri);
}

EXPORT(int) FsRenMovFileW(const WCHAR*OldName,const WCHAR*NewName,BOOL Move,BOOL OverWrite,RemoteInfoStruct*ri) {
	if (lstrlenW(OldName)<pluginrootlen+2 || lstrlenW(NewName)<pluginrootlen+2)
		return FS_FILE_NOTFOUND;

	int err=ProgressProcT(OldName,NewName,0);
	if (err)
		return FS_FILE_USERABORT;

	if (Move) {
		if (OverWrite)
			DeleteFileT(NewName+pluginrootlen);
		if (MoveFileT(OldName+pluginrootlen,NewName+pluginrootlen))
			return FS_FILE_OK;
	} else {
		if (CopyFileT(OldName+pluginrootlen,NewName+pluginrootlen,!OverWrite))
			return FS_FILE_OK;
	}
	switch(GetLastError()) {
		case ERROR_FILE_NOT_FOUND:
		case ERROR_PATH_NOT_FOUND:
		case ERROR_ACCESS_DENIED:
			return FS_FILE_NOTFOUND;
		case ERROR_FILE_EXISTS:
			return FS_FILE_EXISTS;
		default:
			return FS_FILE_WRITEERROR;
	}
	err=ProgressProcT(OldName,NewName,100);
	if (err)
		return FS_FILE_USERABORT;
}

int WINAPI FsGetFileW(WCHAR*RemoteName,WCHAR*LocalName,int CopyFlags,RemoteInfoStruct*ri) {
    int err;
	BOOL ok,OverWrite,Resume,Move;

	OverWrite=CopyFlags & FS_COPYFLAGS_OVERWRITE;
	Resume=CopyFlags & FS_COPYFLAGS_RESUME;
	Move=CopyFlags & FS_COPYFLAGS_MOVE;

	if (Resume)
		return FS_FILE_NOTSUPPORTED;

	if (lstrlenW(RemoteName)<pluginrootlen+2)
		return FS_FILE_NOTFOUND;
	
	err=ProgressProcT(RemoteName,LocalName,0);
	if (err)
		return FS_FILE_USERABORT;
	if (Move) {
		if (OverWrite)
			DeleteFileT(LocalName);
		ok=MoveFileT(RemoteName+pluginrootlen,LocalName);
	} else
		ok=CopyFileT(RemoteName+pluginrootlen,LocalName,!OverWrite);
	
	if (ok) {
		ProgressProcT(RemoteName,LocalName,100);
		return FS_FILE_OK;
	} else {
		err=GetLastError();
		switch (err) {
		case 2:
		case 3:
		case 4:
		case 5:
			return FS_FILE_NOTFOUND;
		case ERROR_FILE_EXISTS:
			return FS_FILE_EXISTS;
		default:
			return FS_FILE_READERROR;
		}
	}
}

int WINAPI FsGetFile(const char*RemoteName,const char*LocalName,int CopyFlags,RemoteInfoStruct*ri) {
 WCHAR RemoteNameW[LONGPATH_MAX],LocalNameW[LONGPATH_MAX];
 return FsGetFileW(
	unicopy(RemoteNameW,elemof(RemoteNameW),RemoteName),
	unicopy(LocalNameW,elemof(LocalNameW),LocalName),
	CopyFlags,ri);
}

#endif

EXPORT(int) FsPutFileW(const WCHAR*local,const WCHAR*name,int flags) {
 if (flags&FS_COPYFLAGS_MOVE) return FS_FILE_NOTSUPPORTED;
 char n[260];
 fs::Entry*e=pathToNode(name,eMedia|eDir,n);
 if (!e) return FS_FILE_WRITEERROR;
 fs::Entry*c;
 for(c=e->sub;c;c=c->next) {
  if (c->type==eFile && !(strcmp(c->name,n))) break;
 }
 if (c) flags|=FS_COPYFLAGS_EXISTS_SAMECASE;
 else{
  for(c=e->sub;c;c=c->next) {
   if (c->type==eFile && !(strcasecmp(c->name,n))) break;
  }
  if (c) flags|=FS_COPYFLAGS_EXISTS_DIFFERENTCASE;
 }
 if (c && !(flags&FS_COPYFLAGS_OVERWRITE)) return FS_FILE_EXISTS;
 HANDLE h=CreateFileW(local,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,0,0);
 if (h==INVALID_HANDLE_VALUE) return FS_FILE_NOTFOUND;
 const Xml*x=Xml::locate(e);
 if (!c) {
  c=x->newNode(e);
  c->type=eFile;
 }
 c->name=x->dup(n);
 *((DWORD*)&c->size)=GetFileSize(h,((DWORD*)&c->size)+1);
 FILETIME ft,ft2;
 if (GetFileTime(h,0,0,&ft)) {
  FileTimeToLocalFileTime(&ft,&ft2);
  FileTimeToDosDateTime(&ft2,((WORD*)c->ftLastMod)+1,(WORD*)c->ftLastMod);
 }
// TODO: MP3-Tags auslesen
 CloseHandle(h);
 return FS_FILE_OK;
}

EXPORT(int) FsPutFile(const char*local,const char*name,int flags) {
 WCHAR localW[260],nameW[260];
 return FsPutFileW(
	unicopy(localW,elemof(localW),local),
	unicopy(nameW,elemof(nameW),name),
	flags);
}

EXPORT(BOOL) FsDeleteFileW(const WCHAR*name) {
 fs::Entry*e=pathToNode(name,eMedia|eDir|eFile);
 if (!e) return FALSE;
 if (e->type!=eFile) return FALSE;
 e->unlink();
 return TRUE;
}

EXPORT(BOOL) FsDeleteFile(const char*name) {
 WCHAR nameW[260];
 return FsDeleteFileW(unicopy(nameW,elemof(nameW),name));
}

EXPORT(BOOL) FsRemoveDirW(const WCHAR*name) {
 fs::Entry*n=pathToNode(name,eMedia|eDir);
 if (!n) return FALSE;
 if (n->type==eCatalog) {// TODO: Katalog aus Liste lschen, nicht den Katalog lschen
// Geht nicht, weil TC erst mal alle Unterverzeichnisse lschen will
 }else n->unlink();
 return TRUE;
}

EXPORT(BOOL) FsRemoveDir(const char*name) {
 WCHAR nameW[LONGPATH_MAX];
 return FsRemoveDirW(unicopy(nameW,elemof(nameW),name));
}

EXPORT(BOOL) FsSetAttrW(const WCHAR*name,int attr) {
 fs::Entry*e=pathToNode(name,eDir|eFile);
 if (!e) return FALSE;
 e->attr=attr;
 return TRUE;
}

EXPORT(BOOL) FsSetAttr(const char*name,int attr) {
 WCHAR nameW[260];
 return FsSetAttrW(unicopy(nameW,elemof(nameW),name),attr);
}

EXPORT(BOOL) FsSetTimeW(const WCHAR*name,const FILETIME*c,const FILETIME*a,const FILETIME*m) {
 fs::Entry*e=pathToNode(name,eCatalog|eMedia|eDir|eFile);
 if (!e) return FALSE;
 if (m) {
  FILETIME ft;
  FileTimeToLocalFileTime(m,&ft);
  FileTimeToDosDateTime(&ft,((WORD*)e->ftLastMod)+1,(WORD*)e->ftLastMod);
  //++dirty
 }
 return TRUE;
}

EXPORT(BOOL) FsSetTime(const char*name,const FILETIME*c,const FILETIME*a,const FILETIME*m) {
 WCHAR nameW[260];
 return FsSetTimeW(unicopy(nameW,elemof(nameW),name),c,a,m);
}

#if 0
EXPORT(void) FsStatusInfo(char* RemoteDir,int InfoStartEnd,int InfoOperation)
{
	// This function may be used to initialize variables and to flush buffers
	
/*	char text[LONGPATH_MAX];

	if (InfoStartEnd==FS_STATUS_START)
		strcpy(text,"Start: ");
	else
		strcpy(text,"End: ");
	
	switch (InfoOperation) {
	case FS_STATUS_OP_LIST:
		strcat(text,"Get directory list");
		break;
	case FS_STATUS_OP_GET_SINGLE:
		strcat(text,"Get single file");
		break;
	case FS_STATUS_OP_GET_MULTI:
		strcat(text,"Get multiple files");
		break;
	case FS_STATUS_OP_PUT_SINGLE:
		strcat(text,"Put single file");
		break;
	case FS_STATUS_OP_PUT_MULTI:
		strcat(text,"Put multiple files");
		break;
	case FS_STATUS_OP_RENMOV_SINGLE:
		strcat(text,"Rename/Move/Remote copy single file");
		break;
	case FS_STATUS_OP_RENMOV_MULTI:
		strcat(text,"Rename/Move/Remote copy multiple files");
		break;
	case FS_STATUS_OP_DELETE:
		strcat(text,"Delete multiple files");
		break;
	case FS_STATUS_OP_ATTRIB:
		strcat(text,"Change attributes of multiple files");
		break;
	case FS_STATUS_OP_MKDIR:
		strcat(text,"Create directory");
		break;
	case FS_STATUS_OP_EXEC:
		strcat(text,"Execute file or command line");
		break;
	case FS_STATUS_OP_CALCSIZE:
		strcat(text,"Calculate space occupied by a directory");
		break;
	case FS_STATUS_OP_SEARCH:
		strcat(text,"Search for file names");
		break;
	case FS_STATUS_OP_SEARCH_TEXT:
		strcat(text,"Search for text in files");
		break;
	case FS_STATUS_OP_SYNC_SEARCH:
		strcat(text,"Search files for sync comparison");
		break;
	case FS_STATUS_OP_SYNC_GET:
		strcat(text,"download files during sync");
		break;
	case FS_STATUS_OP_SYNC_PUT:
		strcat(text,"Upload files during sync");
		break;
	case FS_STATUS_OP_SYNC_DELETE:
		strcat(text,"Delete files during sync");
		break;
	default:
		strcat(text,"Unknown operation");
	}
	if (InfoOperation != FS_STATUS_OP_LIST)   // avoid recursion due to re-reading!
		MessageBox(0,text,RemoteDir,0);
*/
}
#endif

// TODO: Laden der XML-Datei vermeiden wenn nur das Katalog-Icon geladen werden soll
EXPORT(int) FsExtractCustomIconW(WCHAR*name,int flags,HICON&ico) {
 const fs::Entry*n=pathToNode(name,eMedia|eDir|eFile);
 if (!n) return FS_ICON_USEDEFAULT;
 debug("FsExtractCustomIconW(%S,%d)\n",name,flags);
 if (n->type==eCatalog) {
  ico=LoadIcon(hInstance,MAKEINTRESOURCE(1));
  return FS_ICON_EXTRACTED;
 }
 if (n->type==eMedia) {
  ico=LoadIcon(hInstance,MAKEINTRESOURCE(2));
  return FS_ICON_EXTRACTED;
 }
 return FS_ICON_USEDEFAULT;
}

int WINAPI FsExtractCustomIcon(char*name,int flags,HICON&ico) {
 WCHAR nameW[260];
 return FsExtractCustomIconW(unicopy(nameW,elemof(nameW),name),flags,ico);
}

void WINAPI FsSetDefaultParams(FsDefaultParamStruct* dps) {
 lstrcpyn(inifilename,dps->DefaultIniName,elemof(inifilename));
 char buf[4000],*p=buf;
 GetPrivateProfileSection("cdcat1",buf,sizeof buf,inifilename);
// Als UTF-8 interpretieren
 LoadedFile*lf=&loadedFiles[0];
 for (int l;l=strlen(p);p+=l+1,lf++) {
  char*q=StrChr(p,'=');
  if (!q) continue;
  size_t l1=q-p;
  if (l1>sizeof lf->displayname-1) l1=sizeof lf->displayname-1;
  memcpy(lf->displayname,p,l1);
  lf->displayname[l1]=0;
  l1 = l-l1;	// Lnge Dateiname inklusive '\0'
  if (l1>sizeof lf->filename) l1=sizeof lf->filename;
  memcpy(lf->filename,q+1,l1);
 }
 loadedFiles.size(int(lf-&loadedFiles[0]));
}
#if 0
BOOL WINAPI FsLinksToLocalFiles() {
 return true;
}

BOOL WINAPI FsGetLocalName(char* RemoteName,int maxlen) {
 if (strlen(RemoteName)<pluginrootlen+2) return false;
 MoveMemory (RemoteName,RemoteName+pluginrootlen,strlen(RemoteName+pluginrootlen)+1);
 return true;
}

BOOL WINAPI FsGetLocalNameW(WCHAR* RemoteName,int maxlen) {
 if (lstrlenW(RemoteName)<pluginrootlen+2) return false;
 MoveMemory(RemoteName,RemoteName+pluginrootlen,2*lstrlenW(RemoteName+pluginrootlen)+2);
 return true;
}

/**************************************************************************************/
/*********************** content plugin = custom columns part! ************************/
/**************************************************************************************/

#define fieldcount 6
char* fieldnames[fieldcount]={
	"size","creationdate","writedate","accessdate","size-delayed","size-ondemand"};

int fieldtypes[fieldcount]={
	ft_numeric_64,ft_datetime,ft_datetime,ft_datetime,ft_numeric_64,ft_numeric_64};

char* fieldunits_and_multiplechoicestrings[fieldcount]={
	"bytes|kbytes|Mbytes|Gbytes","","","","bytes|kbytes|Mbytes|Gbytes","bytes|kbytes|Mbytes|Gbytes"};

int fieldflags[fieldcount]={
    contflags_substsize,contflags_edit,contflags_substdatetime,contflags_edit,contflags_substsize,contflags_substsize | contflags_edit};

int sortorders[fieldcount]={-1,-1,-1,-1,-1,-1};


int WINAPI FsContentGetSupportedField(int FieldIndex,char* FieldName,char* Units,int maxlen)
{
	if (FieldIndex<0 || FieldIndex>=fieldcount)
		return ft_nomorefields;
	lstrcpyn(FieldName,fieldnames[FieldIndex],maxlen);
	lstrcpyn(Units,fieldunits_and_multiplechoicestrings[FieldIndex],maxlen);
	return fieldtypes[FieldIndex];
}

int WINAPI FsContentGetValueT(BOOL unicode,WCHAR* FileName,int FieldIndex,int UnitIndex,void* FieldValue,int maxlen,int flags) {
 HANDLE fh;
 ULONGLONG filesize;

 if (lstrlenW(FileName+pluginrootlen)<=3) return ft_fileerror;
 if (flags & CONTENT_DELAYIFSLOW) {
  if (FieldIndex==4) return ft_delayed;
  if (FieldIndex==5) return ft_ondemand;
 }
 WIN32_FIND_DATAW W;
 fh=FindFirstFileT(FileName+pluginrootlen,W);
 if (fh!=INVALID_HANDLE_VALUE) {
  FindClose(fh);
  switch (FieldIndex) {
   case 0:  // "size"
   case 4:  // "size-delayed"
   case 5:  // "size-ondemand"
   filesize=W.nFileSizeHigh;
   filesize=(filesize<<32) + W.nFileSizeLow;
   switch (UnitIndex) {
    case 1:
    filesize>>=10;
    break;
    case 2:
    filesize>>=20;
    break;
    case 3:
    filesize>>=30;
    break;
   }
   *(ULONGLONG*)FieldValue=filesize;
   break;
   case 1:  // "creationdate"
   *(FILETIME*)FieldValue=W.ftCreationTime;
   break;
   case 2:  // "writedate"
   *(FILETIME*)FieldValue=W.ftLastWriteTime;
   break;
   case 3:  // "accessdate"
   *(FILETIME*)FieldValue=W.ftLastAccessTime;
   break;
   default: return ft_nosuchfield;
  }
 }else return ft_fileerror;
 return fieldtypes[FieldIndex];  // very important!
}

int WINAPI FsContentGetValueW(WCHAR* FileName,int FieldIndex,int UnitIndex,void* FieldValue,int maxlen,int flags) {
 return FsContentGetValueT(true,FileName,FieldIndex,UnitIndex,FieldValue,maxlen,flags);
}

int WINAPI FsContentGetValue(char* FileName,int FieldIndex,int UnitIndex,void* FieldValue,int maxlen,int flags) {
 WCHAR FileNameW[LONGPATH_MAX];
 return FsContentGetValueT(false,unicopy(FileNameW,elemof(FileNameW),FileName),
	FieldIndex,UnitIndex,FieldValue,maxlen,flags);
}

int WINAPI FsContentGetSupportedFieldFlags(int FieldIndex)
{
	if (FieldIndex==-1)
		return contflags_substmask | contflags_edit;
	else if (FieldIndex<0 || FieldIndex>=fieldcount)
		return 0;
	else
		return fieldflags[FieldIndex];
}

int WINAPI FsContentGetDefaultSortOrder(int FieldIndex)
{
	if (FieldIndex<0 || FieldIndex>=fieldcount)
		return 1;
	else 
		return sortorders[FieldIndex];
}

BOOL WINAPI FsContentGetDefaultView(char* ViewContents,char* ViewHeaders,char* ViewWidths,char* ViewOptions,int maxlen)
{
	lstrcpyn(ViewContents,"[=<fs>.size.bkM2]\\n[=tc.size.bkM2]",maxlen);  // separated by backslash and n, not new lines!
	lstrcpyn(ViewHeaders,"fs-size\\ntc-size",maxlen);  // titles in ENGLISH also separated by backslash and n, not new lines!
	lstrcpyn(ViewWidths,"148,23,-35,-35",maxlen);
	lstrcpyn(ViewOptions,"-1|0",maxlen);  // auto-adjust-width, or -1 for no adjust | horizonal scrollbar flag
	return true;
}

int WINAPI FsContentSetValueW(WCHAR* FileName,int FieldIndex,int UnitIndex,int FieldType,void* FieldValue,int flags)
{
 int retval=ft_nomorefields;
 FILETIME oldcreationtime,newcreationtime;
 FILETIME *p1,*p2,*FieldTime;
 SYSTEMTIME st1,st2;
 HANDLE f;

 if (!FileName) return ft_nosuchfield;    // indicates end of operation -> may be used to flush data
 if (FieldIndex<0 || FieldIndex>=fieldcount) return ft_nosuchfield;
 if (!(fieldflags[FieldIndex]&1)) return ft_nosuchfield;
 switch (FieldIndex) {
  case 1:  // "creationdate"
  case 3:  // "accessdate"
			FieldTime=(FILETIME*)FieldValue;
			p1=NULL;p2=NULL;
			if (FieldIndex==1)
				p1=&oldcreationtime;
			else
				p2=&oldcreationtime;

			f= CreateFileT(FileName+pluginrootlen,	
                      GENERIC_READ|GENERIC_WRITE, // Open for reading+writing
                      0,                      // Do not share
                      NULL,                   // No security
                      OPEN_EXISTING,          // Existing file only
                      FILE_ATTRIBUTE_NORMAL,  // Normal file
                      NULL);
			if (flags & setflags_only_date) {
				GetFileTime(f,p1,p2,NULL);
				FileTimeToLocalFileTime(&oldcreationtime,&newcreationtime);
				FileTimeToSystemTime(&newcreationtime,&st2);
				FileTimeToLocalFileTime(FieldTime,&newcreationtime);
				FileTimeToSystemTime(&newcreationtime,&st1);
				st1.wHour=st2.wHour;
				st1.wMinute=st2.wMinute;
				st1.wSecond=st2.wSecond;
				st1.wMilliseconds=st2.wMilliseconds;
				SystemTimeToFileTime(&st1,&newcreationtime);
				LocalFileTimeToFileTime(&newcreationtime,&oldcreationtime);
			} else
				oldcreationtime=*FieldTime;
			if (!SetFileTime(f,p1,p2,NULL))
				retval=ft_fileerror;
			CloseHandle(f);
			break;
 }
 return retval;
}

int WINAPI FsContentSetValue(char* FileName,int FieldIndex,int UnitIndex,int FieldType,void* FieldValue,int flags) {
 WCHAR FileNameW[LONGPATH_MAX];
 return FsContentSetValueW(unicopy(FileNameW,elemof(FileNameW),FileName),
	FieldIndex,UnitIndex,FieldType,FieldValue,flags);
}

void WINAPI FsContentPluginUnloading() {
	// If you do something in a background thread, you may
	// wait in this function until the thread has finished
	// its work to prevent Total Commander from closing!
	// MessageBox(0,"fsplugin unloading!","Test",0);
}
#endif